#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <itdb.h>

enum {LIST_NONE,LIST_TRACKS,LIST_ALBUMS};
int listtype = LIST_NONE;
enum {MATCH_NONE,MATCH_ALBUM};
int matchtype = MATCH_NONE;
char * argtomatch;
int m3umode=0;
int sortmode=0;
int hcsmode=0;

gint compareTracks(gconstpointer gcpa, gconstpointer gcpb) {
    Itdb_Track const * a = gcpa, * b = gcpb;

    if (a->cd_nr-b->cd_nr) return a->cd_nr-b->cd_nr;
    if (a->track_nr-b->track_nr) return a->track_nr-b->track_nr;

    return strcmp(a->title,b->title);
}

gint compareAlbums(gconstpointer gcpa, gconstpointer gcpb) {
    Itdb_Track const * a = gcpa, * b = gcpb;

    return strcmp(a->album,b->album);
}

int parseArguments(int argc, char ** argv) {
    int i;

    // parse args
    for (i=2; i<argc; i++) {
	/* --list */
	if (!strcmp(argv[i],"--list")) {
	    i++;
	    if (i>=argc) {
		fprintf(stderr,"no argument to --list\n");
		return 1;
	    }
	    if (!strcmp(argv[i],"albums")) listtype=LIST_ALBUMS;
	    else if (!strcmp(argv[i],"tracks")) listtype=LIST_TRACKS;
	    else {
		fprintf(stderr,"bad argument to --list: %s\n",argv[i]);
		return 1;
	    }
	} else if (!strcmp(argv[i],"--match")) {
	    i++;
	    if (i>=argc) {
		fprintf(stderr,"no argument to --match\n");
		return 1;
	    }
	    if (!strcmp(argv[i],"album")) {
		matchtype=MATCH_ALBUM;
		i++;
		if (i>=argc) {
		    fprintf(stderr,"no argument to --match album\n");
		    return 1;
		}
		argtomatch=argv[i];
	    } 
	    else {
		fprintf(stderr,"bad argument to --match: %s\n",argv[i]);
		return 1;
	    }
	} else if (!strcmp(argv[i],"--m3u")) {
	    m3umode=1;
	} else if (!strcmp(argv[i],"--sort")) {
	    sortmode=1;
	} else if (!strcmp(argv[i],"--hcs")) {
	    hcsmode=1;
	}
	else {
	    fprintf(stderr,"bad argument: %s\n",argv[i]);
	    return 1;
	}
    }

    return 0;
}

GList * matchAndSort(GList * tracklist) {
    GList * iterator=NULL;
    GList * sorted=NULL;

	iterator = g_list_first(tracklist);
	while (iterator) {
	    Itdb_Track * track = iterator->data;
	    if ((matchtype == MATCH_NONE) ||
		(matchtype == MATCH_ALBUM && !strcmp(argtomatch,track->album))) {
		if (sortmode) {
		    sorted = g_list_insert_sorted(sorted, track, compareTracks);
		} else {
		    sorted = g_list_append(sorted, track);
		}
	    }

	    iterator = g_list_next(iterator);
	}
    return sorted;
}

void display(GList * sorted, FILE * output) {
    GList * iterator=NULL;
    int i;
    char path[0x100];

    if (m3umode) {
	fprintf(output,"#EXTM3U\n");
    }
    iterator = g_list_first(sorted);
    while (iterator) {
	Itdb_Track * track = iterator->data;
	if (listtype == LIST_ALBUMS) {
     	    fprintf(output,"%s\n",track->album);
	} else if (listtype == LIST_TRACKS) {
    	    if (m3umode) {
		for (i=0;track->ipod_path[i];i++) {
		    if (track->ipod_path[i]==':') path[i]='/';
		    else path[i]=track->ipod_path[i];
		}
		path[i]='\0';
		fprintf(output,"#EXTINF:%d,%s\n",track->tracklen/1000,track->title);
		fprintf(output,"%s\n",path);
	    } else fprintf(output,"%s\n",track->title);
	}
	iterator = g_list_next(iterator);
    }
}

int main(int argc, char ** argv) {
    GError * error = NULL;
    Itdb_iTunesDB * db = NULL;
    GList * sorted = NULL;
    char path[0x100];

    fprintf(stderr,"iTunes DB to m3u convertor\nUses libgpod for DB reading\n\n");

    if (argc<3) {
	fprintf(stderr,"usage: %s iTunesDB [--list albums|tracks] [--match album Albumname] [--sort] [--m3u] [--hcs]\n",argv[0]);
	return 1;
    }

    db = itdb_parse_file(argv[1],&error);

    if (!db) {
	fprintf(stderr,"error opening %s\n",argv[1]);
	return 1;
    }

    if (parseArguments(argc,argv)) {
	itdb_free(db); db=NULL;
	return 1;
    }

    if (hcsmode) {
	GList * albumlist;
	GList * lastalbum = NULL;
	GList * iterator = NULL;

	listtype = LIST_ALBUMS;
	matchtype = MATCH_NONE;

	albumlist = matchAndSort(db->tracks);
	albumlist = g_list_sort(albumlist,compareAlbums);
	iterator = g_list_first(albumlist);
	while (iterator) {
	    if (!lastalbum || compareAlbums(lastalbum->data,iterator->data)) {
		FILE * outlist;
		int i;
		argtomatch = ((Itdb_Track *)iterator->data)->album;
		sprintf(path,"%s.m3u",argtomatch);
		for (i=0;path[i];i++) {
		    switch (path[i]) {
			case '"':
			case '/':
			case '*':
			case '?':
			case '<':
			case '>':
			case '|':
			case ':':
			    path[i]='_';
			    break;
			default:
			    break;
		    }
		}
			
		outlist=fopen(path,"w");
		if (!outlist) {
		    fprintf(stderr,"error opening %s for writing\n",path);
		    itdb_free(db); db=NULL;
		    return 1;
		}
		printf("writing %s\n",path);
		matchtype = MATCH_ALBUM;
		listtype = LIST_TRACKS;
		m3umode=1;
		sortmode=1;

		sorted = matchAndSort(db->tracks);
		display(sorted,outlist);
		g_list_free(sorted); sorted=NULL;
		fclose(outlist); outlist=NULL;
	    }
	    lastalbum=iterator;
	    iterator = g_list_next(iterator);
	}
    } else {
    if (listtype != LIST_NONE) {
	sorted = matchAndSort(db->tracks);
	display(sorted,stdout);
	g_list_free(sorted); sorted=NULL;
    } else {
	fprintf(stderr,"what's the point of not listing?\n");
	itdb_free(db); db=NULL;
	return 1;
    }
    }

    itdb_free(db); db=NULL;
    return 0;
}
